/*
 * _____________________________________________________________________________
 * Copyright 2020-2023 InvenSense Inc. All rights reserved.
 *
 * This software, related documentation and any modifications thereto 
 * (collectively "Software") is subject to InvenSense and its licensors' 
 * intellectual property rights under U.S. and international copyright
 * and other intellectual property rights laws.
 *
 * InvenSense and its licensors retain all intellectual property and proprietary
 * rights in and to the Software and any use, reproduction, disclosure or 
 * distribution of the Software without an express license agreement from 
 * InvenSense is strictly prohibited.
 *
 * EXCEPT AS OTHERWISE PROVIDED IN A LICENSE AGREEMENT BETWEEN THE PARTIES, THE 
 * SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  EXCEPT AS OTHERWISE 
 * PROVIDED IN A LICENSE AGREEMENT BETWEEN THE PARTIES, IN NO EVENT SHALL
 * INVENSENSE BE LIABLE FOR ANY DIRECT, SPECIAL, INDIRECT, INCIDENTAL, OR 
 * CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF 
 * USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER 
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THE SOFTWARE.
 * _____________________________________________________________________________
 */

/*! \file app_config.h */
#ifndef APP_CONFIG_H
#define APP_CONFIG_H

#include <invn/soniclib/soniclib.h>

/* Header files for optional sensor firmware plug-ins referenced below */
#include <invn/soniclib/sensor_fw/ch101/ch101_gpr_narrow.h>
#include <invn/soniclib/sensor_fw/ch101/ch101_gpr_sr.h>
#include <invn/soniclib/sensor_fw/ch101/ch101_gpr_sr_narrow.h>
#include <invn/soniclib/sensor_fw/ch101/ch101_gprmt.h>


/*======================== Sensor Firmware Selection =========================*/

/* Select sensor firmware to use 
 *   The sensor firmware type is specified during the call to ch_init(), by
 *   giving the name (address) of the firmware initialization function that will
 *   be called.  The CHIRP_SENSOR_FW_INIT_FUNC symbol is used to specify the 
 *   init routine for the sensor firmware to be used.
 *
 *   Uncomment ONE of the following lines to use that sensor firmware type.
 *   You must choose a firmware type that is appropriate for the sensor model 
 *   you are using (CH101 or CH201).  
 *
 *   To use a different sensor firmware type than those listed here (for 
 *   example, a new distribution from Chirp), simply define 
 *   CHIRP_SENSOR_FW_INIT_FUNC to equal the name of the init routine for 
 *   the new firmware.
 *
 * 	 Short Range Firmware:
 *   CH101 sensor firmware with "sr" in the name, e.g. ch101_gpr_sr, is 
 *   optimized for short range. The short range firmware has 4 times the 
 *   resolution, but only 1/4 the maximum range.  If you use these options, you 
 *   should redefine the CHIRP_SENSOR_MAX_RANGE_MM symbol, below, to match the
 *   required sensing distance.
 */

	/* CH101 GPR - general purpose rangefinding, standard range */
#define	 CHIRP_SENSOR_FW_INIT_FUNC	ch101_gpr_init

	/* CH101 GPR NARROW - general purpose rangefinding, narrow FoV */
// #define	 CHIRP_SENSOR_FW_INIT_FUNC	ch101_gpr_narrow_init

	/* CH101 GPR SR - general purpose rangefinding, short range */
// #define	 CHIRP_SENSOR_FW_INIT_FUNC	ch101_gpr_sr_init

	/* CH101 GPR SR NARROW - general purpose, short range, narrow FoV */
//  #define	 CHIRP_SENSOR_FW_INIT_FUNC	ch101_gpr_sr_narrow_init

	/* CH101 GPRMT - multi threshold */
// #define	 CHIRP_SENSOR_FW_INIT_FUNC	ch101_gprmt_init

	/* CH201 GPRMT - multi threshold */
// #define	 CHIRP_SENSOR_FW_INIT_FUNC	ch201_gprmt_init


/*=========================== Sensor Configuration ===========================*/

/* Operating mode for the sensor(s)
 *	 These two values define the mode in which the sensor(s) will operate.  
 *
 *	 The CHIRP_FIRST_SENSOR_MODE value specifies the mode for the first sensor
 *	 (lowest numbered) that is present.  If only one sensor is attached, this
 *	 value must be either CH_MODE_TRIGGERED_TX_RX or CH_MODE_FREERUN.
 *
 *	 The CHIRP_OTHER_SENSOR_MODE value specifies the mode for all other sensors
 *	 that are present.
 *
 * 	 For typical Pitch-Catch operation using two or more sensors, set 
 * 	 CHIRP_FIRST_SENSOR_MODE to CH_MODE_TRIGGERED_TX_RX and set 
 * 	 CHIRP_OTHER_SENSOR_MODE to CH_MODE_TRIGGERED_RX_ONLY.
 */
#define CHIRP_FIRST_SENSOR_MODE		CH_MODE_TRIGGERED_TX_RX
#define CHIRP_OTHER_SENSOR_MODE		CH_MODE_TRIGGERED_RX_ONLY

/* Maximum detection range for the sensor
 * This value will determine how long the sensor "listens" for an ultrasound 
 * signal.  Note that the maximum possible range will vary depending on sensor 
 * model (CH101 vs. CH201) and sensor firmware type.  If the value specified 
 * here is greater than the maximum possible range, the maximum possible range 
 * will be used.
 */
#define	CHIRP_MAX_RANGE_MM				(750)	/* maximum range, in mm */

/* Static target rejection range
 * This value specifies if static target rejection (STR) will be used.  If
 * CHIRP_SENSOR_STATIC_RANGE is non-zero, STR will be enabled and will apply
 * to the specified number of samples at the beginning of a measurement.
 */
#define	CHIRP_STATIC_REJECT_SAMPLES		(0)	/* static target rejection sample 
											   range, in samples (0=disabled) */

/* Receive sensor pre-triggering
 * This value specifies if receive (rx) sensor pre-triggering will be used.  
 * This setting only applies if more than one sensor is used, and one or more 
 * sensor is operating in CH_MODE_TRIGGERED_RX_ONLY mode.  
 *
 * Receive pre-triggering improves performance in pitch-catch operation at 
 * short distances, by triggering the receive-only sensor(s) slightly before 
 * the transmitting sensor.  However, this setting will reduce maximum range of 
 * rx-only sensors approximately 200mm, relative to the 
 * CHIRP_SENSOR_MAX_RANGE_MM setting, above.
 *
 * Set CHIRP_RX_PRETRIGGER_ENABLE to non-zero to enable rx pretriggering,
 * or zero to disable.
 */

#define CHIRP_RX_PRETRIGGER_ENABLE	 1

/* Target detection thresholds (if supported)
 * These definitions set the sensor's target detection thresholds, only for 
 * sensors using multi-threshold firmware (e.g. ch101_gprmt or ch201_gprmt).
 *
 * These thresholds specify how large a signal must be received, and at what 
 * point in the measurement, for the sensor to indicate that a target was 
 * detected and calculate the range.
 *
 * Each threshold consists of a starting sample number within the measurement 
 * and the corresponding amplitude value that must be reached.  A threshold 
 * extends until the starting sample number of the next threshold, if any.
 */
#define CHIRP_THRESH_0_START	(0)       /* threshold 0 - must start at zero */
#define CHIRP_THRESH_0_LEVEL	(5000)

#define CHIRP_THRESH_1_START	(26)      /* threshold 1 */
#define CHIRP_THRESH_1_LEVEL	(2000)

#define CHIRP_THRESH_2_START	(39)      /* threshold 2 */
#define CHIRP_THRESH_2_LEVEL	(800)

#define CHIRP_THRESH_3_START	(56)      /* threshold 3 */
#define CHIRP_THRESH_3_LEVEL	(400)

#define CHIRP_THRESH_4_START	(79)      /* threshold 4 */
#define CHIRP_THRESH_4_LEVEL	(250)

#define CHIRP_THRESH_5_START	(89)      /* threshold 5 */
#define CHIRP_THRESH_5_LEVEL	(175)


/*============================ Application Timing ============================*/

/* Define how often the application will get a new sample from the sensor(s) 
 *   This macro defines the sensor measurement interval, in milliseconds.  
 *
 *   For sensors in triggered mode (CH_MODE_TRIGGERED_TX_RX or 
 *   CH_MODE_TRIGGERED_RX_ONLY), the application will use a periodic timer to 
 *   trigger a sensor measurement each time this period elapses.
 *
 *   For sensors in free-running mode (CH_MODE_FREERUN), the application will
 *   set this period as the sensor's internal sample interval.
 */

#define	MEASUREMENT_INTERVAL_MS (100)


/*==================  Application Storage for Sensor Data ====================*/

/* Define how many samples per measurement are expected by this application
 *   The following macro is used to allocate array storage in the "chirp_data_t"
 *   structure, defined in main.c.  That structure contains arrays for 
 *   individual sample data values (I/Q pairs) that describe the raw samples 
 *   within an ultrasound measurement.  
 *
 *   This the space in the chirp_data_t structure is only allocated if
 *	 OUTPUT_AMP_DATA_CSV or OUTPUT_IQ_DATA is defined (see below).
 *
 *   Because a CH201 sensor has the highest number of samples in each 
 *   measurement, the full CH201 sample count is used here to guarantee space 
 *   in all configurations.  You may reduce this value to the actual number of 
 *   samples needed for your use.
 */

#define APP_DATA_MAX_SAMPLES  CH201_MAX_NUM_SAMPLES


/*================  Build Options for Sample Data Handling ==================*/

/* The following build options control if and how the full set of values for 
 * all internal samples within an ultrasound measurement will be read and 
 * displayed.  This data is separate from the standard range and simple target 
 * amplitude values that are normally output.
 *
 * The measurement data can be read and displayed either as net amplitude values
 * or as individual I and Q components.
 * 
 * Note that reading the full data set is not required for most basic 
 * sensing applications - the reported range value, possibly combined with the 
 * simple target amplitude value, is typically all that is required.  However, 
 * the full set of sample values may be read and analyzed for more advanced 
 * sensing or data capture needs.
 *
 * Set OUTPUT_AMP_DATA_CSV to 1 (non-zero) to enable output of the amplitude 
 * data via the serial port, as a series of comma separated values all on one 
 * line.  This allows easy cut/paste into a spreadsheet program to generate 
 * an "A-Scan" chart.  This kind of summary graph can be very helpful to 
 * understand what the sensor is observing.
 *
 * Set OUTPUT_IQ_DATA to 1 (non-zero) to enable output of the I/Q pair values.
 * Each I/Q sample appears as a comma separated pair of (Q,I) values on its own 
 * line.  Having the separate I and Q values allows more complex analysis of 
 * signal phase, etc.
 *
 * OUTPUT_AMP_DATA_CSV and OUTPUT_IQ_DATA cannot be used together.
 */
#define OUTPUT_AMP_DATA_CSV  	(0)		/* if non-zero, output amp data */
#define OUTPUT_IQ_DATA			(0)		/* if non-zero, output I/Q data */

/* Blocking vs. non-blocking I/O
 * By default, this application will read sample data in blocking mode
 * (i.e. READ_DATA_NONBLOCKING is zero by default).  The amplitude or I/Q data 
 * will be read from the device and placed in the corresponding data array 
 * field in the application's chirp_data structure.  Because the data is read 
 * in blocking mode, the calls to ch_get_iq_data() will not return until all 
 * the data has actually been copied from the sensor over I2C.
 *
 * However, if READ_DATA_NONBLOCKING is non-zero, the sample data will be read 
 * in non-blocking mode. The ch_get_iq_data() calls will return immediately, 
 * and a separate callback function will be called to notify the application 
 * when the read operation is complete.
 * 
 * If neither OUTPUT_AMP_DATA_CSV nor OUTPUT_IQ_DATA is selected (non-zero),
 * this option has no effect.
 */
#define READ_DATA_NONBLOCKING  	(0)		/* if non-zero, use non-blocking mode */


/*======================  Application Display Options ========================*/

/* These symbols control what values are displayed for each measurement. */
#define	DISPLAY_AMP_VALUE		(1)		/* if non-zero, show target amplitude */
#define	DISPLAY_SAMPLE_NUM		(1)		/* if non-zero, show target sample num*/


#endif /* APP_CONFIG_H */
