/***********************************************************************
 * Hello Chirp! - an example application for ultrasonic sensing
 *
 * This project is designed to be your first introduction to using
 * the SonicLib API to control ultrasonic sensors in an embedded C
 * application.
 *
 * It configures connected CH101 or CH201 sensors, sets up a measurement
 * timer, and triggers the sensors each time the timer expires.
 * On completion of each measurement, it reads out the sensor data and
 * prints it over the console serial port.
 *
 * The settings used to configure the sensors are defined in
 * the app_config.h header file.
 *
 ***********************************************************************/

/*
 Copyright 2020-2023, Invensense, Inc. All rights reserved.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.
 */

/* Includes */
#include <stdio.h>
#include <asf.h>
#include <invn/soniclib/soniclib.h>			// Chirp SonicLib sensor API definitions
#include "chirp_board_config.h"	// required header with basic device counts etc.
#include "app_config.h"
#include "app_version.h"
#include <invn/soniclib/chirp_bsp.h>			// board support package function definitions
#include "ultrasound_display_config_info.h"

/* Bit flags used in main loop to check for completion of sensor I/O.  */
#define DATA_READY_FLAG		(1 << 0)
#define IQ_READY_FLAG		(1 << 1)

/* Check build option dependencies */
#if (OUTPUT_AMP_DATA_CSV && OUTPUT_IQ_DATA)
#error Cannot specify multiple output options at the same time (OUTPUT_AMP_xxx or OUTPUT_IQ_xxx).
#endif

#if (OUTPUT_AMP_DATA_CSV || OUTPUT_IQ_DATA)
#define READ_IQ_DATA		1			// need to read I/Q for amplitude data or I/Q output
#else
#define READ_IQ_DATA		0
#endif

/* chirp_data_t - Structure to hold measurement data for one sensor
 *   This structure is used to hold the data from one measurement cycle from
 *   a sensor.  The data values include the measured range, the ultrasonic
 *   signal amplitude, the number of valid samples (I/Q data pairs) in the
 *   measurement, and (optionally) the full amplitude data and/or raw I/Q data
 *   from the measurement.
 *
 *  The format of this data structure is specific to this application, so
 *  you may change it as desired.
 *
 *  A "chirp_data[]" array of these structures, one for each possible sensor,
 *  is declared in the main.c file.  The sensor's device number is
 *  used to index the array.
 */
typedef struct
{
	uint32_t range;						// from ch_get_range()
	uint16_t amplitude;					// from ch_get_amplitude()
	uint16_t num_samples;				// from ch_get_num_samples()
#if READ_IQ_DATA
	ch_iq_sample_t	iq_data[APP_DATA_MAX_SAMPLES];
										// from ch_get_iq_data()
#endif
} chirp_data_t;

/* Array of structs to hold measurement data, one for each possible device */
chirp_data_t chirp_data[CHIRP_MAX_NUM_SENSORS];

/* Array of ch_dev_t device descriptors, one for each possible device */
static ch_dev_t chirp_devices[CHIRP_MAX_NUM_SENSORS];

/* Configuration structure for group of sensors */
ch_group_t chirp_group;

/* Detection threshold settings - only for sensors using multi-threshold firmware
 *   (e.g. ch101_gprmt or ch201_gprmt).
 *   Each threshold entry includes the starting sample number & threshold level.
 */
ch_thresholds_t chirp_detect_thresholds = {
	.threshold = {
		{CHIRP_THRESH_0_START, CHIRP_THRESH_0_LEVEL},      /* threshold 0 */
		{CHIRP_THRESH_1_START, CHIRP_THRESH_1_LEVEL},      /* threshold 1 */
		{CHIRP_THRESH_2_START, CHIRP_THRESH_2_LEVEL},      /* threshold 2 */
		{CHIRP_THRESH_3_START, CHIRP_THRESH_3_LEVEL},      /* threshold 3 */
		{CHIRP_THRESH_4_START, CHIRP_THRESH_4_LEVEL},      /* threshold 4 */
		{CHIRP_THRESH_5_START, CHIRP_THRESH_5_LEVEL},      /* threshold 5 */
	}
};

/* Task flag word
 *   This variable contains the DATA_READY_FLAG, IQ_READY_FLAG and
 *   DATA_READY_FLAG bit flags that are set in I/O processing routines.  The
 *   flags are checked in the main() loop and, if set, will cause an
 *   appropriate handler function to be called to process sensor data.
 */
volatile uint32_t taskflags = 0;

/* Device tracking variables
 *   These are bit-field variables which contain a separate bit assigned to
 *   each (possible) sensor, indexed by the device number.  The active_devices
 *   variable contains a bit pattern describing which ports have active
 *   sensors connected.  The data_ready_devices variable is set bit-by-bit
 *   as sensors interrupt, indicating they have completed a measurement
 *   cycle.  The two variables are compared to determine when all active
 *   devices have interrupted.
 */
static uint32_t active_devices;
static uint32_t data_ready_devices;

/* Number of connected sensors */
static uint8_t num_connected_sensors = 0;

/* Number of sensors that use external triggering to start measurement */
static uint8_t num_triggered_sensors = 0;

#if (READ_IQ_DATA && READ_DATA_NONBLOCKING)
/* Count of non-blocking I/Q reads queued */
static uint8_t	num_io_queued = 0;
#endif

/* Forward declarations */
static uint8_t configure_sensors(ch_group_t * grp_ptr, uint8_t num_ports);
static void sensor_int_callback(ch_group_t *grp_ptr, uint8_t dev_num, ch_interrupt_type_t int_type);
static void periodic_timer_callback(void);
static uint8_t handle_data_ready(ch_group_t *grp_ptr);
#if (OUTPUT_IQ_DATA || OUTPUT_AMP_DATA_CSV)
static void display_sensor_data(chirp_data_t * sensor_data);
#endif

#if READ_IQ_DATA
static uint8_t read_iq_data(ch_dev_t *dev_ptr);
#endif

#if READ_DATA_NONBLOCKING
static void io_complete_callback(ch_group_t *grp_ptr);
#endif

#if (READ_IQ_DATA && READ_DATA_NONBLOCKING)
static uint8_t handle_iq_data_done(ch_group_t *grp_ptr);
#endif


/* main() - entry point and main loop
 *
 * This function contains the initialization sequence for the application
 * on the board, including system hardware initialization, sensor discovery
 * and configuration, callback routine registration, and timer setup.  After
 * the initialization sequence completes, this routine enters an infinite
 * loop that will run for the remainder of the application execution.
 */

int main(void)
{
	ch_group_t *grp_ptr = &chirp_group;
	uint8_t chirp_error = 0;
	ch_version_t  lib_ver;
	uint8_t num_ports;
	uint8_t dev_num;
	ch_dev_t *dev_ptr;

	/* Initialize board hardware functions
	 *   This call to the board support package (BSP) performs all necessary
	 *   hardware initialization for the application to run on this board.
	 *   This includes setting up memory regions, initializing clocks and
	 *   peripherals (including I2C and serial port), and any processor-specific
	 *   startup sequences.
	 *
	 *   The chbsp_board_init() function also initializes fields within the
	 *   sensor group descriptor, including number of supported sensors and
	 *   the RTC clock calibration pulse length.
	 */
	chbsp_board_init(grp_ptr);

	printf("\n\nHello Chirp! - SonicLib Example Application for CHx01 Sensors\n");
	printf("    Compile time:  %s %s\n", __DATE__, __TIME__);
	printf("    Version: %u.%u.%u", APP_VERSION_MAJOR, APP_VERSION_MINOR, APP_VERSION_REV);
	ch_get_version(&lib_ver);
	printf("    SonicLib version: %u.%u.%u\n", lib_ver.major, lib_ver.minor, lib_ver.rev);
	printf("\n");

	/* Get the number of (possible) sensor devices on the board
	 *   Set by the BSP during chbsp_board_init()
	 */
	num_ports = ch_get_num_ports(grp_ptr);

	/* Initialize sensor descriptors.
	 *   This loop initializes each (possible) sensor's ch_dev_t descriptor,
	 *   although we don't yet know if a sensor is actually connected.
	 *
	 *   The call to ch_init() specifies the sensor descriptor, the sensor group
	 *   it will be added to, the device number within the group, and the sensor
	 *   firmware initialization routine that will be used.  (The sensor
	 *   firmware selection effectively specifies whether it is a CH101 or
	 *   CH201 sensor, as well as the exact feature set.)
	 */
	printf("Initializing sensor(s)... ");

	for (dev_num = 0; dev_num < num_ports; dev_num++) {
		dev_ptr = &(chirp_devices[dev_num]);	// init struct in array
		/* Init device descriptor
		 *   Note that this assumes all sensors will use the same sensor
		 *   firmware.  The CHIRP_SENSOR_FW_INIT_FUNC symbol is defined in
		 *   app_config.h and is used for all devices.
		 *
		 *   However, it is possible for different sensors to use different
		 *   firmware images, by specifying different firmware init routines
		 *   when ch_init() is called for each.
		 */
		chirp_error |= ch_init(dev_ptr, grp_ptr, dev_num, CHIRP_SENSOR_FW_INIT_FUNC);
	}

	/* Start all sensors.
	 *   The ch_group_start() function will search each port (that was
	 *   initialized above) for a sensor. If it finds one, it programs it (with
	 *   the firmware specified above during ch_init()) and waits for it to
	 *   perform a self-calibration step.  Then, once it has found all the
	 *   sensors, ch_group_start() completes a timing reference calibration by
	 *   applying a pulse of known length to the sensor's INT line.
	 */
	if (chirp_error == 0) {
		printf("starting group... ");
		chirp_error = ch_group_start(grp_ptr);
	}

	if (chirp_error == 0) {
		printf("OK\n");
	} else {
		printf("FAILED: %d\n", chirp_error);
	}
	printf("\n");

	/* Get and display the initialization results for each connected sensor.
	 *   This loop checks each device number in the sensor group to determine
	 *   if a sensor is actually connected.  If so, it makes a series of
	 *   function calls to get different operating values, including the
	 *   operating frequency, clock calibration values, and firmware version.
	 */
	printf("Sensor\tType \t   Freq\t\t B/W  \t RTC Cal \tFirmware\n");

	for (dev_num = 0; dev_num < num_ports; dev_num++) {
		dev_ptr = ch_get_dev_ptr(grp_ptr, dev_num);

		if (ch_sensor_is_connected(dev_ptr)) {
			printf("%d\tCH%d\t %lu Hz\t%u Hz\t%u@%ums\t%s\n", dev_num,
					ch_get_part_number(dev_ptr),
					ch_get_frequency(dev_ptr),
					ch_get_bandwidth(dev_ptr),
					ch_get_rtc_cal_result(dev_ptr),
					ch_get_rtc_cal_pulselength(dev_ptr),
					ch_get_fw_version_string(dev_ptr));
		}
	}
	printf("\n");

	/* Register callback function to be called when Chirp sensor interrupts */
	ch_io_int_callback_set(grp_ptr, sensor_int_callback);

	/* Register callback function called when non-blocking I/Q readout completes
	 *   Note, this callback will only be used if READ_DATA_NONBLOCKING is defined
	 *   as non-zero to enable non-blocking I/Q readout in this application.
	 */
#if READ_DATA_NONBLOCKING
	ch_io_complete_callback_set(grp_ptr, io_complete_callback);
#endif // READ_DATA_NONBLOCKING

	/* Configure each sensor with its operating parameters
	 *   Initialize a ch_config_t structure with values defined in the
	 *   app_config.h header file, then write the configuration to the
	 *   sensor using ch_set_config().
	 */
	printf ("Configuring sensor(s)...\n");
	chirp_error |= configure_sensors(grp_ptr, num_ports);

	/* Initialize interrupts and timer based on whether triggering is used */
	if (num_triggered_sensors == 0) {
		/* All sensors are in freerun mode, so configure INT line input to receive 
		 * interrupt from sensor */
		chdrv_int_group_set_dir_in(grp_ptr);			// set INT line as input
		chdrv_int_group_interrupt_enable(grp_ptr);		// enable interrupt

	} else {
		/* Initialize the periodic timer we'll use to trigger the measurements.
		 *   This function initializes a timer that will interrupt every time it
		 *   expires, after the specified measurement interval.  The function also
		 *   registers a callback function that will be called from the timer
		 *   handler when the interrupt occurs.  The callback function will be
		 *   used to trigger a measurement cycle on the group of sensors.
		 */
		printf("Initializing sample timer for %dms interval... ", MEASUREMENT_INTERVAL_MS);

		chbsp_periodic_timer_init(MEASUREMENT_INTERVAL_MS, periodic_timer_callback);

		/* Enable timer interrupt and start timer to trigger sensor sampling */
		chbsp_periodic_timer_irq_enable();
		chbsp_periodic_timer_start();

		printf("OK\n");
	}

	printf("Starting measurements\n");
	
	/* Enter main loop
	 *   This is an infinite loop that will run for the remainder of the system
	 *   execution.  The processor is put in a low-power sleep mode between
	 *   measurement cycles and is awakened by interrupt events.
	 *
	 *   The interrupt may be the periodic timer (set up above), a data-ready
	 *   interrupt from a sensor, or the completion of a non-blocking I/O
	 *   operation.  This loop will check flags that are set during the
	 *   callback functions for the data-ready interrupt and the non-blocking
	 *   I/O complete.  Based on the flags that are set, this loop will call
	 *   the appropriate routines to handle and/or display sensor data.
	 */
	while (1) {
		/*
		 * Put processor in light sleep mode if there are no pending tasks, but
		 * never turn off the main clock, so that interrupts can still wake
		 * the processor.
		 *
		 * Disable interrupts before going to the sleep mode. It assumes the 
		 * interrupts can be re-enabled as part of the MCU sleep mechanism 
		 */
#ifdef __arm__
		__disable_irq();				// disable interrupts to check int flags
#endif
		if (taskflags == 0) {

			chbsp_proc_sleep();			// put processor in low-power sleep mode
			/* We only continue here after an interrupt wakes the processor */

		} else {
#ifdef __arm__
			__enable_irq();				// re-enable interrupts after checking flags
#endif			
		}

		/* Check for sensor data-ready interrupt(s) */
		if (taskflags & DATA_READY_FLAG) {

			/* All sensors have interrupted - handle sensor data */
			taskflags &= ~DATA_READY_FLAG;		// clear flag
			handle_data_ready(grp_ptr);			// read and display measurement
		}

#if (READ_IQ_DATA && READ_DATA_NONBLOCKING)
		/* Start any pending non-blocking I/Q reads */
		if (num_io_queued != 0) {
			ch_io_start_nb(grp_ptr);
			num_io_queued = 0;
		}

		/* Check for non-blocking I/Q read complete */
		if (taskflags & IQ_READY_FLAG) {

			/* All non-blocking I/Q readouts have completed */
			taskflags &= ~IQ_READY_FLAG;		// clear flag
			handle_iq_data_done(grp_ptr);		// display I/Q data
		}
#endif
	}	// end while(1) infinite loop

}	// end main()

/* 
 * configure_sensors() - apply configuration settings
 * 
 * This function performs the detailed configuration of the sensor.
 * Values defined in the app_config.h file are used here as parameters to
 * individual SonicLib API functions.
 */
static uint8_t configure_sensors(ch_group_t * grp_ptr, uint8_t num_ports)
{
	uint8_t dev_num;
	ch_dev_t *dev_ptr;
	uint8_t chirp_error;
	ch_config_t dev_config;

	chirp_error = 0;
	num_connected_sensors = 0;
	active_devices = 0;

	for (dev_num = 0; dev_num < num_ports; dev_num++) {
		dev_ptr = ch_get_dev_ptr(grp_ptr, dev_num);

		if (!chirp_error && ch_sensor_is_connected(dev_ptr)) {

			num_connected_sensors++;			// count one more connected sensor
			active_devices |= (1 << dev_num);	// add to active device bit mask

			/* Set mode */
			if (num_connected_sensors == 1) {	// if this is the first sensor
				dev_config.mode = CHIRP_FIRST_SENSOR_MODE;
			} else {
				dev_config.mode = CHIRP_OTHER_SENSOR_MODE;
			}
			
			/* Init config structure with values from app_config.h */
			dev_config.max_range       = CHIRP_MAX_RANGE_MM;
			dev_config.static_range    = CHIRP_STATIC_REJECT_SAMPLES;

			/* If sensor will be free-running, set internal sample interval */
			if (dev_config.mode == CH_MODE_FREERUN) {
				dev_config.sample_interval = MEASUREMENT_INTERVAL_MS;
			} else {
				dev_config.sample_interval = 0;
				num_triggered_sensors++;	// count one more triggered sensor
			}

			/* Set detection thresholds if sensor firmware supports them */
			if (ch_get_num_thresholds(dev_ptr) > 1) {
				/* Set pointer to struct containing detection thresholds */
				dev_config.thresh_ptr = &chirp_detect_thresholds;	
			} else {
				dev_config.thresh_ptr = 0;							
			}

			/* Enable sensor interrupt if using free-running mode 
			 *   Note that interrupt is automatically enabled if using 
			 *   triggered modes.
			 */
			if ((!chirp_error) && (dev_config.mode == CH_MODE_FREERUN)) {
				chdrv_int_set_dir_in(dev_ptr);
				chdrv_int_interrupt_enable(dev_ptr);
			}

			/* Apply sensor configuration */
			if (!chirp_error) {
				chirp_error = ch_set_config(dev_ptr, &dev_config);
			}

			/* Read back and display config settings */
			if (!chirp_error) {
				ultrasound_display_config_info(dev_ptr);
			} else {
				printf("Device %d: Error during ch_set_config()\n", dev_num);
			}

			/* Turn on an LED to indicate device connected */
			if (!chirp_error) {
				chbsp_led_on(dev_num);
			}
		}	// end if (!chirp_error && ch_sensor_is_connected)
	}	// end for (dev_num < num_ports)

	/* Enable receive sensor pre-triggering, if specified */
	ch_set_rx_pretrigger(grp_ptr, CHIRP_RX_PRETRIGGER_ENABLE);

	return chirp_error;
}

/*
 * periodic_timer_callback() - periodic timer callback routine
 *
 * This function is called by the periodic timer interrupt when the timer
 * expires.  It sets a flag to that will be checked in the main() loop.
 *
 * If the sensor is operating in triggered mode this function calls ch_group_trigger()
 * during each execution.
 *
 * This function is registered by the call to chbsp_periodic_timer_init()
 * in main().
 */

static void periodic_timer_callback(void)
{
	if (num_triggered_sensors > 0) {
		ch_group_trigger(&chirp_group);
	}
}

/*
 * sensor_int_callback() - sensor interrupt callback routine
 *
 * This function is called by the board support package's interrupt handler for
 * the sensor's INT line every time that the sensor interrupts.  The device
 * number parameter, dev_num, is used to identify the interrupting device
 * within the sensor group.  (Generally the device number is same as the port
 * number used in the BSP to manage I/O pins, etc.)
 *
 * Each time this function is called, a bit is set in the data_ready_devices
 * variable to identify the interrupting device.  When all active sensors have
 * interrupted (found by comparing with the active_devices variable), the
 * DATA_READY_FLAG is set.  That flag will be detected in the main() loop.
 *
 * This callback function is registered by the call to ch_io_int_callback_set()
 * in main().
 */
static void sensor_int_callback(ch_group_t *grp_ptr, uint8_t dev_num,
								__attribute__((unused)) ch_interrupt_type_t int_type)
{
	ch_dev_t *dev_ptr = ch_get_dev_ptr(grp_ptr, dev_num);

	data_ready_devices |= (1 << dev_num);	// add to data-ready bit mask

	chbsp_led_toggle(dev_num);				// use LED to indicate activity on this sensor

	if (data_ready_devices == active_devices) {
		/* All active sensors have interrupted after performing a measurement */
		data_ready_devices = 0;

		/* Set data-ready flag - it will be checked in main() loop */
		taskflags |= DATA_READY_FLAG;

		/* Disable sensor interrupts unless in free-running mode
		 *   They will automatically be re-enabled by the next ch_group_trigger()
		 */
		if (num_triggered_sensors == 0) {			// if all sensors in freerun mode
			chdrv_int_set_dir_in(dev_ptr);				// set INT line as input
			chdrv_int_group_interrupt_enable(grp_ptr);	// leave interrupt enabled
		} else {
			chdrv_int_group_interrupt_disable(grp_ptr);	// leave interrupt disabled
		}
	}
}

#if READ_DATA_NONBLOCKING
/*
 * io_complete_callback() - non-blocking I/O complete callback routine
 *
 * This function is called by SonicLib's I2C DMA handling function when all
 * outstanding non-blocking I/Q readouts have completed.  It simply sets a flag
 * that will be detected and handled in the main() loop.
 *
 * This callback function is registered by the call to
 * ch_io_complete_callback_set() in main().
 *
 *  Note: This callback is only used if READ_DATA_NONBLOCKING is defined as
 *  non-zero to select non-blocking I/Q readout in this application.
 */
static void io_complete_callback(ch_group_t __attribute__((unused)) *grp_ptr)
{
	taskflags |= IQ_READY_FLAG;
}
#endif // READ_DATA_NONBLOCKING

/*
 * handle_data_ready() - get and display data from all sensors
 *
 * This routine is called from the main() loop after all sensors have
 * interrupted. It shows how to read the sensor data once a measurement is
 * complete.  This routine always reads out the range and amplitude, and
 * optionally will read out the amplitude data or raw I/Q for all samples
 * in the measurement.
 *
 * See the comments in app_config.h for information about the amplitude data
 * and I/Q readout build options.
 *
 */
static uint8_t handle_data_ready(ch_group_t *grp_ptr)
{
	uint8_t dev_num;
	int num_samples = 0;
	uint8_t ret_val = 0;
	uint32_t range;
	uint16_t amplitude;
	ch_range_t range_type;

	/* Read and display data from each connected sensor
	 *   This loop will write the sensor data to this application's "chirp_data"
	 *   array.  Each sensor has a separate chirp_data_t structure in that
	 *   array, so the device number is used as an index.
	 */
	for (dev_num = 0; dev_num < ch_get_num_ports(grp_ptr); dev_num++) {
		ch_dev_t *dev_ptr = ch_get_dev_ptr(grp_ptr, dev_num);

		if (ch_sensor_is_connected(dev_ptr)) {

			 /*  For sensors in transmit/receive mode, report one-way echo
			 *   range.  For sensors in receive-only mode, report direct
			 *   one-way range from transmitting sensor.
			 */
			if (ch_get_mode(dev_ptr) == CH_MODE_TRIGGERED_RX_ONLY) {
				range_type = CH_RANGE_DIRECT;
			} else {
				range_type = CH_RANGE_ECHO_ONE_WAY;
			}

			printf("Dev %d:  ", dev_num);

			range = ch_get_range(dev_ptr, range_type);
			chirp_data[dev_num].range = range;


			if (range != CH_NO_TARGET) {
				/* Target object was successfully detected (range available) */

				 /* Get the new amplitude value - it's only updated if range
				  * was successfully measured.  */
				amplitude = ch_get_amplitude(dev_ptr);
				chirp_data[dev_num].amplitude = amplitude;

				printf("Range: %0.1f mm ", (float) range/32.0f);

				if (DISPLAY_SAMPLE_NUM) {
					uint16_t num_mm = (range / 32);

					/* ch_mm_to_samples() assumes num_mm is one-way (1/2 round-trip time-of-flight) */
					if (range_type != CH_RANGE_ECHO_ONE_WAY) {
						num_mm /= 2;		// mm distance was entire time-of-flight, so convert
					}
					printf("(sample %u) ", ch_mm_to_samples(dev_ptr, num_mm));
				}

				if (DISPLAY_AMP_VALUE) {
					printf("amp=%5u ", chirp_data[dev_num].amplitude);
				}

			} else {
				/* No target object was detected - no range value */
				chirp_data[dev_num].amplitude = 0;  /* no updated amplitude */

				printf("        no target found  ");
			}

			/* Store number of active samples in this measurement */
			num_samples = ch_get_num_samples(dev_ptr);
			chirp_data[dev_num].num_samples = num_samples;

#if READ_IQ_DATA
			/* Optionally read raw I/Q values for all samples */
			read_iq_data(dev_ptr);

#if !(READ_DATA_NONBLOCKING)
			/* Output sensor data immediately if read in blocking mode */
			display_sensor_data(&chirp_data[dev_num]);
#endif

#endif // READ_IQ_DATA
			
			printf("\n");
		}
	}

	return ret_val;
}

#if READ_IQ_DATA
/* read_iq_data() - Read full I/Q data from device into buffer or queue read
 * This function is used to obtain the full I/Q data from the sensor following a
 * measurement.  I/Q data may be displayed directly, or it may be converted to
 * amplitude values for output.
 *
 * Depending on build options, this routine will either read the I/Q data
 * immediately, before returning, or it will queue a non-blocking read of the data.
 * If READ_DATA_NONBLOCKING is zero (default), this function will read the data
 * from the sensor into the application's "chirp_data" structure for this device
 * before returning.
 *
 * If READ_DATA_NONBLOCKING is non-zero, the I/Q read operation will be queued
 * and this routine will return immediately.  A callback routine will be called
 * when the read operation is complete.  The callback routine must have been
 * registered using the ch_io_complete_callback_set() function.
 */
uint8_t read_iq_data(ch_dev_t *dev_ptr)
{
	uint16_t 	start_sample = 0;
	uint8_t 	error = 0;
	uint16_t 	num_samples = ch_get_num_samples(dev_ptr);
	uint8_t		dev_num = ch_get_dev_num(dev_ptr);

#if (READ_DATA_NONBLOCKING == 0)
	/* Reading I/Q data in normal, blocking mode */
	error = ch_get_iq_data(dev_ptr, chirp_data[dev_num].iq_data,
								start_sample, num_samples, CH_IO_MODE_BLOCK);

	if (error) {
		printf("     Error reading %d I/Q samples", num_samples);
	}

#else
	/* Reading I/Q data in non-blocking mode - queue a read operation */
	printf("     queuing %d I/Q samples... ", num_samples);

	error = ch_get_iq_data(dev_ptr, chirp_data[dev_num].iq_data,
								start_sample, num_samples, CH_IO_MODE_NONBLOCK);

	if (!error) {
		num_io_queued++;		// record a pending non-blocking read
		printf("OK");
	} else {
		printf("**ERROR**");
	}
#endif  // defined(READ_DATA_BLOCKING)

	return error;
}
#endif  // READ_IQ_DATA


#if (READ_IQ_DATA && READ_DATA_NONBLOCKING)
/*
 * handle_iq_data_done() - handle raw I/Q data from a non-blocking read
 *
 * This function is called from the main() loop when a queued non-blocking
 * readout of the raw I/Q data has completed for all sensors.  The data will
 * have been placed in this application's "chirp_data" array, in the
 * chirp_data_t structure for each sensor, indexed by the device number.
 *
 * By default, this function takes no action on the I/Q data, except to display
 * the number of samples that were read from the device.
 */
static uint8_t handle_iq_data_done(ch_group_t *grp_ptr)
{
	int        dev_num;
	ch_dev_t * dev_ptr;

	for (dev_num = 0; dev_num < ch_get_num_ports(grp_ptr); dev_num++) {
		dev_ptr = ch_get_dev_ptr(grp_ptr, dev_num);

		if (ch_sensor_is_connected(dev_ptr)) {

			chirp_data[dev_num].num_samples = ch_get_num_samples(dev_ptr);
			printf ("Read %d samples from device %d:", chirp_data[dev_num].num_samples, dev_num);

#if (OUTPUT_IQ_DATA || OUTPUT_AMP_DATA_CSV)
			display_sensor_data(&chirp_data[dev_num]);
#endif
		}
	}

	return 0;
}
#endif  // (READ_IQ_DATA && READ_DATA_NONBLOCKING)

#if (OUTPUT_IQ_DATA || OUTPUT_AMP_DATA_CSV)
/* 
 * display_sensor_data() - display full sample data from a measurement
 *
 * This function displays the complete set of sample values from a measurement.  Either
 * the amplitude or raw I/Q values may be output, based on the definitions of
 * OUTPUT_AMP_DATA_CSV and OUTPUT_IQ_DATA.
 */
static void display_sensor_data(chirp_data_t * sensor_data)
{
	uint16_t num_samples;
	uint16_t count;
	ch_iq_sample_t * iq_ptr = (ch_iq_sample_t *)&sensor_data->iq_data[0];

	/* Output IQ/Amp values in CSV format, one sample (pair or value) per line */
	num_samples = (sensor_data->num_samples <= APP_DATA_MAX_SAMPLES) ? sensor_data->num_samples : APP_DATA_MAX_SAMPLES;

	for (count = 0; count < num_samples; count++) {
#if OUTPUT_AMP_DATA_CSV
		uint16_t sample_amp = ch_iq_to_amplitude(iq_ptr);		// calculate sample amplitude from I/Q
		printf(", %u", sample_amp);								// output on one line, comma separated
#elif OUTPUT_IQ_DATA
		printf("\n%d,%d", iq_ptr->q, iq_ptr->i);				// output one I/Q pair per line
#endif
		iq_ptr++;		// next sample
	}
}
#endif // (OUTPUT_IQ_DATA || OUTPUT_AMP_DATA_CSV)
