/*
 Copyright 2016-2023, InvenSense, Inc.  All rights reserved.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.

 */
#include <asf.h>
#include "uart_hal.h"
#include "chirp_smartsonic.h"

int8_t uart_hal_init(uint8_t rts_cts, uint8_t *uart_rx_buffer)
{
	const usart_serial_options_t uart_serial_options = {
		.baudrate = CONF_UART_BAUDRATE,
		.charlength = CONF_UART_CHAR_LENGTH,
		.paritytype = CONF_UART_PARITY,
		.stopbits = CONF_UART_STOP_BITS,
	};

	/* Configure console UART. */
	sysclk_enable_peripheral_clock(ID_FLEXCOM0);
	usart_serial_init(USART0, (usart_serial_options_t *)&uart_serial_options);

	uint32_t mode = (0 == rts_cts) ? US_MR_CHMODE_NORMAL : US_MR_USART_MODE_HW_HANDSHAKING;
	USART0->US_MR = (USART0->US_MR & ~US_MR_USART_MODE_Msk) | mode;

	/* Prepare for reception */
	uart_hal_dma_getc(uart_rx_buffer, 1);

	usart_enable_rx(USART0);
	usart_enable_tx(USART0);

	/* Enable UART IRQ.
	 * Note that this UART INT must have the highest priority in the system to be able to
	 * receive all bytes coming from the FTDI. */
	NVIC_SetPriority(FLEXCOM0_IRQn, INT_PRIO_UART_COMM);
	NVIC_EnableIRQ(FLEXCOM0_IRQn);

	return 0;
}

bool uart_hal_is_connected(void)
{
	// Can't know if USB cable is connected or not. So suppose yes.
	return true;
}

void uart_hal_dma_getc(uint8_t * data, uint32_t len)
{
	pdc_packet_t g_pdc_usart_packet; /* PDC data packet for transfer */
	Pdc *g_p_usart_pdc; /* Pointer to USART PDC register base */

	/* Get pointer to USART PDC register base */
	g_p_usart_pdc = usart_get_pdc_base(USART0);

	/* Initialize PDC data packet for transfer */
	g_pdc_usart_packet.ul_addr = (uint32_t) data;
	g_pdc_usart_packet.ul_size = len;

	/* Configure PDC for data receive */
	pdc_rx_init(g_p_usart_pdc, &g_pdc_usart_packet, NULL);

	/* Enable PDC transfers */
	pdc_enable_transfer(g_p_usart_pdc, PERIPH_PTCR_RXTEN);
	usart_enable_interrupt(USART0, US_IER_ENDRX);
}

void uart_hal_dma_puts(const uint8_t * data, uint32_t len)
{
	/* PDC data packet for transfer */
	pdc_packet_t g_pdc_usart_packet;
	/* Pointer to USART PDC register base */
	Pdc *g_p_usart_pdc;

	/* Get pointer to USART PDC register base */
	g_p_usart_pdc = usart_get_pdc_base(USART0);

	/* Initialize PDC data packet for transfer */
	g_pdc_usart_packet.ul_addr = (uint32_t) data;
	g_pdc_usart_packet.ul_size = len;

	/* Configure PDC for data receive */
	pdc_tx_init(g_p_usart_pdc, &g_pdc_usart_packet, NULL);

	/* Enable PDC transfers */
	pdc_enable_transfer(g_p_usart_pdc, PERIPH_PTCR_TXTEN);
	usart_enable_interrupt(USART0, US_IER_ENDTX);
}

/* Atmel USART IRQ Handler */
void FLEXCOM0_Handler(void)
{
	uint32_t ul_status;

	/* Read USART Status. */
	ul_status = usart_get_status(USART0);

	if((ul_status &  US_CSR_ENDRX ))
	{
		uart_hal_rx_received_callback();
	}
	if((ul_status &  US_CSR_ENDTX ))
	{
		uart_hal_tx_sent_callback();
		usart_disable_interrupt(USART0, US_IER_ENDTX);
	}
}