<#
.SYNOPSIS

Script to create and release package.

.DESCRIPTION

This script creates the soniclib release package using the
CreateSourceCodeRelease.bat script, uploads it to swdpackages, then publishes
the documentation to github pages. This is run by the jeknins package job but
can also be run locally.

.PARAMETER tag
The release tag to assign to the output package. This should match the git tag.

.PARAMETER push_doc
Set to push the documentation to the gh-pages branch, which publishes
it to the website. If not set, then the commit will be made locally, but the
branch will not be pushed. You can check out the local branch and check for
validity.

.PARAMETER public_release
If set, publish the output package to the public github repository.

.PARAMETER public_commit_message
The commit message to publish to the public repo.

.PARAMETER build_library
Run the library build and fail on compilation errors.

.PARAMETER git_user_name
The user name to use when making a git commit. Should not normally be changed
from default.

.PARAMETER git_user_email
The user name to use when making a git commit. Should not normally be changed
from default.

.PARAMETER git_repo_url
The git repo URL for git pushes. Should point to the public github repo for this
project.

.PARAMETER git_ignores
List of files to ignore when making the gh-pages push. This is populated into
a .gitignore file.

.INPUTS

None.

.OUTPUTS

None.

.EXAMPLE

PS> .\jenkins-package.ps1 1.0.0 -push_doc

#>

param(
    [Parameter(Mandatory, HelpMessage="Enter the tag to package:")]
    [string]$tag,
    [Parameter(HelpMessage="Push to gh-pages branch?")]
    [switch]$push_doc,
    [Parameter(HelpMessage="Publish to public github repo?")]
    [switch]$public_release,
    [Parameter(HelpMessage="Enter public commit message:")]
    [string]$public_commit_message="",
    [Parameter(HelpMessage="Build library?")]
    [switch]$build_library,
    [Parameter(HelpMessage="Enter user name (for git commit):")]
    [string]$git_user_name="tdk-opensource",
    [Parameter(HelpMessage="Enter user email (for git commit):")]
    [string]$git_user_email="inv-inv-tdk-opensource@tdk.com",
    [Parameter(HelpMessage="Enter repo url (for git push):")]
    [string]$git_repo_url="git@github.com:tdk-invn-oss/ultrasonic.soniclib.git",
    [Parameter(HelpMessage="Enter gitignores list for gh-pages push:")]
    [string[]]$git_ignores=@("dist","html","release","src","release.log")
)

# Configure names of packages to copy and location to copy in swdpackages
$output_pkg_name = "SonicLib_$tag"
$archive_names = ("$output_pkg_name.zip","$output_pkg_name.exe")
$archive_rel_path = "invn\chirpmicro\soniclib"

function Is-Git-Status-Dirty {
    -not [string]::IsNullOrEmpty((git diff --ignore-space-change 2> $null))
}

function Does-Tag-Exist {
    param([string]$tag)
    git rev-parse "$tag" 2>&1>$null
    return ($LASTEXITCODE -eq 0)
}

function Copy-To-SwdPackages {
    param([string]$local_location,
          [string[]]$archive_names,
          [string]$dist_location)
    try {
        pushd "$local_location"
        ForEach ($archive_name in $archive_names) {
            if (-not (Test-Path "$dist_location/$archive_name")) {
                echo "Copying $archive_name to $dist_location..."
                cp "$archive_name" "$dist_location"
            } else {
                throw "$archive_name already exists in $archive_location!"
                # Below doesn't work because zip creation isn't deterministic
                # apparently.
                # if the files are the same, don't generate an error
                # $hash_new = (Get-FileHash -Path "$archive_name").Hash
                # $hash_old = (Get-FileHash -Path "$dist_location\$archive_name").Hash
                # if ($hash_new -ne $hash_old) {
                #     throw (("$archive_name already exists at $dist_location",
                #             "and contents do not match! Existing hash=$hash_old.",
                #             "New hash=$hash_new.") -join " ")
                # } else {
                #     echo "$archive_name already exists at $dist_location, " `
                #          "but contents are identical. No error will be raised."
                #}
            }
        }
    } finally {
        popd
    }
}

function Update-Github-Pages {
    param([string]$ghpages_branch,
          [string]$html_dir,
          [string]$tag,
          [string]$commit_message,
          [switch]$push_and_tag,
          [switch]$push,
          [string]$tag_prefix="ghpages",
          [string]$remote="",
          [string[]]$gitignores=$null,
          [string]$git_user_name=$null,
          [string]$git_user_email=$null)
    try {
        echo "Stashing changes and switching to the $ghpages_branch branch..."
        git stash
        $commit = git rev-parse HEAD
        if ((-not [string]::IsNullOrEmpty($remote))) {
            git remote add public "$remote"
            $remote = "public"
        } else {
            $remote = "origin"
        }
        # delete all local tags and re-fetch them for the correct remote
        git tag | foreach-object -process { git tag -d $_ } > $null
        git fetch "$remote" 2>&1>$null
        $ghpages_branch_name = "$remote-$ghpages_branch"
        git checkout -B "$ghpages_branch_name" "$remote/$ghpages_branch" 2>&1>$null
        if ($LASTEXITCODE) {
            echo "checkout $ghpages_branch_name failed, creating new orphan branch..."
            git checkout --orphan "$ghpages_branch_name"
            git rm --cached *
            if ($gitignores) {
                $gitignores | Out-File .gitignore -Encoding ASCII
                git add .gitignore
            }
        } else {
            # ensure we have a clean and up to date branch, then update the html files
            echo "Ensuring we have a clean and up to date $ghpages_branch_name branch..."
            git reset --hard "$remote/$ghpages_branch"
        }
        git clean -df
        # ugly, but removes all files tracked by git. We're going to copy everything
        # from the last build, so we don't want any old stuff hanging around. Exclude
        # ignored files, especially html, and .gitignore. The former contains the new
        # build to be copied and the latter is a necessary tracked file that is not
        # regenerated.
        echo "Removing all non-ignored content (preserving .gitignore)..."
        # below loads .gitignore into list, ignoring empty lines
        $ignored_content = (Get-Content ".gitignore") | ? {$_.trim() -ne ""}
        $ignored_content += (".gitignore") # add gitignore itself
        try {
            (ls | select -ExpandProperty fullname) `
                -NotMatch ($ignored_content -join "|") | rm -Recurse
        } catch {
            echo "[WARNING] Removing tracked files failed. Empty repo?"
        }
        echo "Adding tracked content from new build..."
        cp -Recurse -Force "$html_dir\*" .
        if ((Does-Tag-Exist "$tag_prefix`_$tag")) {
            if ((Is-Git-Status-Dirty)) {
                $gitdiff = git --no-pager diff --ignore-space-change 2>$null
                # echo "gitdiff=$gitdiff"
                throw "Tag $tag_prefix`_$tag already exists in remote with different contents!"
            } else {
                echo "Tag already exists and contents matches. Skipping " `
                     "commit, push, and tag steps."
            }
        } else {
            git add .
            if ($git_user_name -and $git_user_email) {
                echo "Committing as $git_user_name <$git_user_email> ..."
                git commit --author="$git_user_name <$git_user_email>" `
                    -m "$commit_message"
            } else {
                git commit -m "$commit_message"
            }
            if ($push -or $push_and_tag) {
                echo "Pushing..."
                git push "$remote" "${ghpages_branch_name}:$ghpages_branch"
            }
            # add a release tag if this is a package job run by jenkins
            if ($push_and_tag) {
                echo "Tagging the doc with $tag_prefix`_$tag..."
                git tag "$tag_prefix`_$tag"
                git push "$remote" "$tag_prefix`_$tag"
            }
        }
    } finally {
        echo "Switching back to original branch and applying stashed changes..."
        git reset --hard "$remote/$ghpages_branch"
        echo "Checking out $commit"
        git checkout "$commit"
        echo "Current branches:"
        git branch
        echo "Applying stashed changes..."
        git stash apply
    }
}

function Update-Github-Repo {
    param([string]$pkg_name,
          [string]$tag,
          [string]$commit_message,
          [string]$url,
          [string]$git_user_name=$null,
          [string]$git_user_email=$null)
    echo "Update-Github-Repo called with pkg_name=$pkg_name, tag=$tag, " `
         "commit_message=$commit_message, url=$url, git_user_name=$git_user_name " `
         "git_user_email=$git_user_email"
    try {
        pushd release
        7z x -o"$pkg_name" "$pkg_name.zip"
        pushd $pkg_name
        git init
        git remote add origin "$url"
        git fetch 2>&1>$null
        # check if tag exists and abort if it does
        if (Does-Tag-Exist "$tag") {
            throw "Tag $tag already exists in remote! Aborting."
        }
        git reset origin/master
        git checkout -b release/$tag
        git add .
        # git config --list --show-origin
        if ($git_user_name -and $git_user_email) {
            echo "Committing as $git_user_name <$git_user_email> ..."
            git commit --author="$git_user_name <$git_user_email>" `
                -m "$commit_message"
        } else {
            git commit -m "$commit_message"
        }
        # git show --summary
        git push origin release/$tag
        # No longer tagging automatically because we will create a PR from
        # release branch to master first.
        # git tag "$tag"
        # git push origin "$tag"
    } finally {
        popd  # output package
        # left commented for debugging purposes
        # rm -Recurse -Force "$pkg_name"
        popd  # release
    }
}

# run the continuous build to check for build errors
if ($build_library) {
    & ".\jenkins-continuous.bat"
    if ($LASTEXITCODE) { exit 1 }
}

# Create the source code release
& ".\CreateSourceCodeRelease.bat" $tag
if ($LASTEXITCODE) { exit 1 }

# set some default values for variables if missing in env or params
$job_name = $env:JOB_NAME
if ($job_name -eq $null) {
    $job_name = "local job by $env:USERNAME"
}

$mbs_dist_repo = $env:MBS_DIST_REPO
if ($mbs_dist_repo -eq $null) {
    $mbs_dist_repo = "\\ussjc-share01\swdpackages"
}

if ($public_commit_message -eq "") {
    $public_commit_message = "Release $tag."
}

echo mbs_dist_repo=$mbs_dist_repo

$is_jenkins_package_job = (($env:USERNAME -eq "jenkins") `
                            -and ($env:JOB_NAME -match "package"))


if ($is_jenkins_package_job) {
    $archive_location = "$mbs_dist_repo\$archive_rel_path"
} else {
    $archive_location = "$pwd\dist\$archive_rel_path"
}
if (-not (Test-Path "$archive_location" -PathType Container)) {
    mkdir "$archive_location"
}
try {
    Copy-To-SwdPackages -local_location "release" `
                        -archive_names $archive_names `
                        -dist_location "$archive_location"
} catch {
    echo "Caught the following exception but continuing anyway..."
    echo $_
    # Don't exit in this case because the public github update won't run. There
    # is some risk of swdpackages and github releases getting out of sync. Need
    # to think about best way to handle this. Maybe just disable swdpackages
    # release at some point...
}

try {
    $UpdateArgs = @{
        ghpages_branch = "gh-pages"
        html_dir = "html"
        tag = "$tag"
        tag_prefix = "ghpages"
        commit_message = "$tag documentation update from $job_name"
        push = $push_doc
        push_and_tag = $is_jenkins_package_job
        gitignores = $git_ignores
    }
    Update-Github-Pages @UpdateArgs
} catch {
    echo $_
    exit 1
}


# now publish to public github if requested
if ($public_release) {
    try {
        Update-Github-Repo -pkg_name "$output_pkg_name" `
                           -tag "$tag" `
                           -commit_message "$public_commit_message" `
                           -url $git_repo_url `
                           -git_user_name $git_user_name `
                           -git_user_email $git_user_email

        $UpdateArgs = @{
            ghpages_branch = "gh-pages"
            html_dir = "html"
            tag = "$tag"
            tag_prefix = "ghpages"
            commit_message = "$tag documentation update from $job_name"
            push = $push_doc
            push_and_tag = $is_jenkins_package_job
            remote = $git_repo_url
            gitignores = $git_ignores
            git_user_name = $git_user_name
            git_user_email = $git_user_email
        }
        Update-Github-Pages @UpdateArgs
    } catch {
        echo $_
        exit 1
    }
}
