/*
 Copyright 2016-2023, InvenSense, Inc.  All rights reserved.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.

 */
#include <stdbool.h>
#include <string.h>

#include "ultrasound_display_config_info.h"
#include <invn/soniclib/soniclib.h>

#define ODR_TO_FREQ_DIV(odr) ( 1 << (7 - odr))

#define SEG_TYPE_TO_STR(segment_type) \
	(segment_type == CH_MEAS_SEG_TYPE_COUNT) ? "Count" :\
	((segment_type == CH_MEAS_SEG_TYPE_RX) ? "RX   " :\
	((segment_type == CH_MEAS_SEG_TYPE_TX) ? "TX   " :\
	((segment_type == CH_MEAS_SEG_TYPE_EOF) ? "EOF  " : "UNKNOWN")))

#define MODE_TO_STR(mode) \
	(mode == CH_MODE_FREERUN) ? "CH_MODE_FREERUN" :\
	((mode == CH_MODE_TRIGGERED_TX_RX) ? "CH_MODE_TRIGGERED_TX_RX" :\
	((mode == CH_MODE_TRIGGERED_RX_ONLY) ? "CH_MODE_TRIGGERED_RX_ONLY" :\
	((mode == CH_MODE_IDLE) ? "CH_MODE_IDLE" : "UNKNOWN")))

static bool is_part_number_icu(ch_dev_t *dev_ptr)
{
	return  ((ch_get_part_number(dev_ptr) == ICU10201_PART_NUMBER) ||
			 (ch_get_part_number(dev_ptr) == ICU20201_PART_NUMBER) ||
			 (ch_get_part_number(dev_ptr) == ICU30201_PART_NUMBER));
}

/*
 * display_config_info() - display the configuration values for a sensor
 *
 * This function displays the current configuration settings for an individual
 * sensor.  The operating mode, maximum range, and static target rejection
 * range (if used) are displayed.
 *
 * For CH201 and ICU202XX sensors, the multiple detection threshold values are also
 * displayed.
 */
uint8_t ultrasound_display_config_info(ch_dev_t *dev_ptr)
{
	ch_config_t sensor_config;
	uint8_t chirp_error = 0;
	uint8_t dev_num = ch_get_dev_num(dev_ptr);

	/* Read configuration values for the device into ch_config_t structure */
	chirp_error = ch_get_config(dev_ptr, &sensor_config);

	if (!chirp_error) {
		if (!is_part_number_icu(dev_ptr)) {
			/* Display sensor number, mode and max range */
			printf("Sensor %u:\tmax_range=%umm (%u samples)\tmode=%s  ", dev_num,
				sensor_config.max_range, ch_get_num_samples(dev_ptr), MODE_TO_STR(sensor_config.mode));

			/* Display static target rejection range, if used */
			if (sensor_config.static_range != 0) {
				printf("static_range=%u samples", sensor_config.static_range);
			}
		}

#ifdef INCLUDE_SHASTA_SUPPORT
		if (is_part_number_icu(dev_ptr)) {
			/* Display measurement configuration */
			ch_meas_info_t meas_info;
			ch_meas_seg_info_t seg_info;

			for (uint8_t meas_num = 0; meas_num < MEAS_QUEUE_MAX_MEAS; meas_num++) {
				ch_meas_get_info(dev_ptr, meas_num, &meas_info);

				if (meas_info.num_segments == 0)
					continue;

				printf("Device %u: Measurement %u Configuration\n", dev_num, meas_num);

				printf("  Total Samples = %u  (%u mm max range) \tRingdown samples = %u\n",
						meas_info.num_rx_samples, ch_samples_to_mm(dev_ptr, meas_info.num_rx_samples),
						meas_info.ringdown_cancel_samples);

				uint8_t odr_freq_div = ODR_TO_FREQ_DIV(meas_info.odr);
				printf("  Active Segments = %u\tRate = CH_ODR_FREQ_DIV_%u\n", meas_info.num_segments, odr_freq_div);

				for (int seg_num = 0; seg_num <= meas_info.num_segments; seg_num++) { /* also display EOF */
					ch_meas_get_seg_info(dev_ptr, meas_num, seg_num, &seg_info);
					printf("   Seg %u  %s ", seg_num, SEG_TYPE_TO_STR(seg_info.type));
					if (seg_info.type == CH_MEAS_SEG_TYPE_RX) {
						printf("%3d sample%s  ", seg_info.num_rx_samples,
								(seg_info.num_rx_samples == 1 ? " ":"s"));
					} else {
						printf("             ");
					}
					printf("%5d cycles  ", seg_info.num_cycles);
					if (seg_info.type == CH_MEAS_SEG_TYPE_TX) {
						printf("Pulse width = %2u  Phase = %u  ", seg_info.tx_pulse_width, seg_info.tx_phase);
					} else if (seg_info.type == CH_MEAS_SEG_TYPE_RX) {
						printf("Gain reduce = %2u  Atten = %u  ", seg_info.rx_gain, seg_info.rx_atten);
					}
					if (seg_info.rdy_int_en) {
						printf("Rdy Int  ");
					}
					if (seg_info.done_int_en) {
						printf("Done Int");
					}
					printf("\n");
				} /* for (seg_num...) */
			} /* for(meas_num...) */
		}
#endif // INCLUDE_SHASTA_SUPPORT

#if defined (CH_NUM_THRESHOLDS)
		/* Get threshold values in structure */
		ch_thresholds_t detect_thresholds;
		chirp_error = ch_get_thresholds(dev_ptr, &detect_thresholds);
		if (!chirp_error) {
			printf("  Detection thresholds:\n");
			for (int thresh_num = 0; thresh_num < CH_NUM_THRESHOLDS; thresh_num++) {
				uint16_t start_sample = detect_thresholds.threshold[thresh_num].start_sample;
				uint16_t start_mm = ch_samples_to_mm(dev_ptr, start_sample);

				if ((thresh_num == 0) || (start_sample != 0)) { /* unused thresholds have start = 0 */
					printf("     %u\tstart sample: %3u  = %4u mm\tlevel: %u",
						thresh_num, start_sample, start_mm,
						detect_thresholds.threshold[thresh_num].level);
#ifdef INCLUDE_SHASTA_SUPPORT
					if (detect_thresholds.threshold[thresh_num].level == CH_THRESH_LEVEL_HOLDOFF) {
						printf(" (Rx Holdoff)");
					}
#endif // INCLUDE_SHASTA_SUPPORT
					printf("\n");
				}
			}
		} /* else : error reading thresholds or thresholds not handled by fw */
#endif
		printf("\n");
	} else {
		printf(" Device %u: ERROR during ch_get_config()\n", dev_num);
	}

	return chirp_error;
}

/* Get and display the initialization results for each connected sensor.
 *   This loop checks each device number in the sensor group to determine
 *   if a sensor is actually connected.  If so, it makes a series of
 *   function calls to get different operating values, including the
 *   operating frequency, clock calibration values, and firmware version.
 */
void ultrasound_display_connected_sensors_measures_info(ch_group_t *grp_ptr)
{
	uint8_t num_ports = ch_get_num_ports(grp_ptr);

	printf("Sensor  Type       Freq      RTC Cal     B/W    CPU Freq   Firmware\r\n");
	for (uint8_t dev_num = 0; dev_num < num_ports; dev_num++) {
		ch_dev_t *dev_ptr = ch_get_dev_ptr(grp_ptr, dev_num);
		if (ch_sensor_is_connected(dev_ptr)) {
			printf("  %u     %s%u  %lu Hz  %u@%ums  %uHz  %0.2fMHz  %s\r\n", dev_num,
				(is_part_number_icu(dev_ptr)) ? "ICU-" : "    CH",
				ch_get_part_number(dev_ptr),
				ch_get_frequency(dev_ptr),
				ch_get_rtc_cal_result(dev_ptr),
				ch_get_rtc_cal_pulselength(dev_ptr),
				ch_get_bandwidth(dev_ptr),
				(ch_get_cpu_frequency(dev_ptr) / 1000000.0f),
				ch_get_fw_version_string(dev_ptr));
		}
	}
	printf("\n");
}

void ultrasound_display_connected_sensors_otp_info(__attribute__((unused)) ch_group_t *grp_ptr)
{
#ifdef INCLUDE_SHASTA_SUPPORT
	uint8_t num_ports = ch_get_num_ports(grp_ptr);
	const char *sensor_id_str;
	ch_mfg_info_t manufacturing_info;

	printf("Sensor\tId\tYear\tWeek\tSite\tProduct\tPackage\tMEMS\tModule\r\n");
	for (uint8_t dev_num = 0; dev_num < num_ports; dev_num++) {
		ch_dev_t *dev_ptr = ch_get_dev_ptr(grp_ptr, dev_num);

		if (ch_sensor_is_connected(dev_ptr)) {
			sensor_id_str = ch_get_sensor_id(dev_ptr);
			if (0 == strcmp(sensor_id_str, "DEFAULT")) {
				printf("%u\tNot programmed\r\n", dev_num);
			} else {
				ch_get_mfg_info(dev_ptr, &manufacturing_info);
				printf("%u\t%s\t%04u\tW%02u\t%u\t%03u\t%u\t%u\t%u\r\n",
					dev_num, sensor_id_str,
					manufacturing_info.mfg_year,manufacturing_info.mfg_week,
					manufacturing_info.mfg_site, manufacturing_info.product_code,
					manufacturing_info.package_code, manufacturing_info.mems_code,
					manufacturing_info.module_code);
			}
		}
	}
#endif // INCLUDE_SHASTA_SUPPORT
}
